from ..models import TimeEntry


class EditorError(Exception): ...


class Editor:
    """Base Editor Service class"""

    def __init__(self, editor_command: str):
        self.editor_command = editor_command

    def edit_entry(self, entry: TimeEntry) -> TimeEntry:
        raise NotImplementedError()



from .time_entry_repo import TimeEntryRepository
from .time_entry_repo_file import TimeEntryRepositoryFile
from .time_entry_repo_orm import TimeEntryRepositoryORM

__all__ = ["TimeEntryRepository", "TimeEntryRepositoryFile", "TimeEntryRepositoryORM"]



import json
from datetime import datetime
from typing import Self

from peewee import CharField, DateTimeField, Model, fn
from playhouse.sqlite_ext import JSONField, SqliteExtDatabase

from ..models import EntryListFilter, TimeEntry
from .time_entry_repo import TimeEntryRepository

db = SqliteExtDatabase(None)


class TimeEntryORM(Model):
    id = CharField(primary_key=True)
    start_time = DateTimeField()
    end_time = DateTimeField(null=True)
    project = CharField()
    comment = CharField()
    tags = JSONField()

    class Meta:
        database = db
        table_name = "time_entries"

    def to_model(self) -> TimeEntry:
        return TimeEntry(
            id=self.id,
            start_time=self.start_time,
            end_time=self.end_time,
            project=self.project,
            comment=self.comment,
            tags=set(self.tags),
        )

    @classmethod
    def create_from_model(cls, entry: TimeEntry) -> Self:
        return cls.create(
            id=entry.id,
            start_time=entry.start_time,
            end_time=entry.end_time,
            project=entry.project,
            comment=entry.comment,
            tags=list(entry.tags),
        )


def json_array_contains(json_array, value):
    return value in json.loads(json_array) if json_array else False


class TimeEntryRepositoryORM(TimeEntryRepository):
    def __init__(self, db_path: str):
        db.init(db_path)
        db.connect()
        db.create_tables([TimeEntryORM])
        db.connection().create_function("json_array_contains", 2, json_array_contains)

    def get_all(self) -> list[TimeEntry]:
        return [entry.to_model() for entry in TimeEntryORM.select().order_by(TimeEntryORM.start_time.asc())]

    def get_by_project(self, project: str) -> list[TimeEntry]:
        return [entry.to_model() for entry in TimeEntryORM.select().where(TimeEntryORM.project == project)]

    def get_entry_by_id(self, id: str) -> TimeEntry:
        try:
            return TimeEntryORM.get(TimeEntryORM.id == id).to_model()
        except TimeEntryORM.DoesNotExist as e:
            raise KeyError("record id not found") from e

    def delete_entry(self, id: str) -> TimeEntry:
        try:
            entry = TimeEntryORM.get(TimeEntryORM.id == id)
            entry.delete_instance()
            return entry.to_model()
        except TimeEntryORM.DoesNotExist as e:
            raise KeyError("record id not found") from e

    def get_active_entry(self):
        return TimeEntryORM.get(TimeEntryORM.end_time.is_null()).to_model()

    def filter(self, *, filter: EntryListFilter | None = None) -> list[TimeEntry]:
        query = TimeEntryORM.select().order_by(TimeEntryORM.start_time.asc())
        if filter:
            if filter.id:
                query = query.where(TimeEntryORM.id.startswith(filter.id))
            if filter.projects:
                query = query.where(TimeEntryORM.project.in_(filter.projects))
            if filter.start_date:
                query = query.where(TimeEntryORM.start_time >= filter.start_date)
            if filter.end_date:
                query = query.where(TimeEntryORM.start_time <= datetime.combine(filter.end_date, datetime.max.time()))
            if filter.tags:
                for tag in filter.tags:
                    query = query.where(fn.json_array_contains(TimeEntryORM.tags, tag))
        return [entry.to_model() for entry in query]

    def save(self, entry: TimeEntry) -> None:
        try:
            orm_entry = TimeEntryORM.get(TimeEntryORM.id == entry.id)
        except TimeEntryORM.DoesNotExist:
            orm_entry = TimeEntryORM.create_from_model(entry)
        orm_entry.save()

    def save_all(self, entries: list[TimeEntry]) -> None:
        with db.atomic():
            data = [entry.model_dump(mode="json") for entry in entries]
            TimeEntryORM.insert_many(data).execute()



import pytest

from ...models import EntryListFilter, TimeEntry
from ..time_entry_repo_orm import TimeEntryRepositoryORM


def test_entry_repo_orm_empty():
    """Test ORM repo with no entries."""
    repo = TimeEntryRepositoryORM(":memory:")

    assert repo.get_all() == []
    assert repo.get_by_project("test") == []
    with pytest.raises(KeyError):
        repo.get_entry_by_id("test")
    assert repo.filter() == []


def test_entry_repo_orm_standard():
    repo = TimeEntryRepositoryORM(":memory:")
    entry = TimeEntry(project="test", start_time="2021-01-01T00:00:00", tags=["tag1", "tag2"])
    repo.save(entry)
    assert repo.get_active_entry() == entry


def test_entry_repo_orm_filter():
    repo = TimeEntryRepositoryORM(":memory:")
    entry1 = TimeEntry(
        project="test",
        start_time="2021-01-01T00:00:00",
        end_time="2021-01-01T01:00:00",
        tags=["tag1", "tag2"],
        comment="test",
    )
    entry2 = TimeEntry(
        project="test2", start_time="2021-01-02T00:00:00", end_time="2021-01-02T01:30:00", tags=["tag2", "tag3"]
    )
    entry3 = TimeEntry(
        project="test3", start_time="2021-01-03T00:00:00", end_time="2021-01-03T02:00:00", tags=["tag1", "tag3"]
    )
    repo.save(entry1)
    repo.save(entry2)
    repo.save(entry3)
    filter1 = EntryListFilter(projects=["test"])
    assert repo.filter(filter=filter1) == [entry1]
    filter2 = EntryListFilter(tags=["tag1"])
    assert repo.filter(filter=filter2) == [entry1, entry3]
    filter3 = EntryListFilter(start_date="2021-01-02")
    assert repo.filter(filter=filter3) == [entry2, entry3]
    assert repo.filter() == [entry1, entry2, entry3]
    filter4 = EntryListFilter(end_date="2021-01-02")
    assert repo.filter(filter=filter4) == [entry1, entry2]
    filter5 = EntryListFilter(id=entry1.id[0:4])
    assert repo.filter(filter=filter5) == [entry1]
    filter6 = EntryListFilter(tags=["blah"])
    assert repo.filter(filter=filter6) == []


def test_entry_repo_orm_delete():
    repo = TimeEntryRepositoryORM(":memory:")
    entry1 = TimeEntry(
        project="test",
        start_time="2021-01-01T00:00:00",
        stop_time="2021-01-01T01:00:00",
        tags=["tag1", "tag2"],
        comment="test",
    )
    repo.save(entry1)
    assert repo.get_all() == [entry1]
    d_entry = repo.delete_entry(entry1.id)
    assert d_entry == entry1
    assert repo.get_all() == []
    with pytest.raises(KeyError):
        repo.delete_entry("test")


def test_entry_repo_orm_save_all():
    repo = TimeEntryRepositoryORM(":memory:")
    entry1 = TimeEntry(
        project="test",
        start_time="2021-01-01T00:00:00",
        stop_time="2021-01-01T01:00:00",
        tags=["tag1", "tag2"],
        comment="test",
    )
    entry2 = TimeEntry(
        project="test2",
        start_time="2021-01-02T00:00:00",
        stop_time="2021-01-02T01:00:00",
        tags=["tag2", "tag3"],
        comment="test2",
    )
    repo.save_all([entry1, entry2])
    assert repo.get_all() == [entry1, entry2]



import os
import subprocess
import tempfile

import rtoml as toml
import ryaml

from ..models import TimeEntry
from . import Editor, EditorError


class EditFormat:
    """Base class for file formats used in editing"""

    def __init__(self, extension: str):
        self.extension = extension

    def __repr__(self):
        return f"{self.__class__.__name__}({self.extension!r})"

    @property
    def suffix(self):
        return f".{self.extension}"


class YAMLFormat(EditFormat):
    """YAML file format"""

    def __init__(self):
        super().__init__("yaml")

    def format_entry_for_edit(self, entry: TimeEntry) -> str:
        """Format a time entry as YAML for editing"""
        entry_dict = entry.model_dump(mode="json")
        return ryaml.dumps(entry_dict)

    def parse_edited_entry(self, content: str) -> TimeEntry:
        """Parse edited YAML content back into a TimeEntry"""
        try:
            data = ryaml.loads(content)
            return TimeEntry(**data)
        except Exception as e:
            raise EditorError(f"Invalid entry format: {str(e)}") from e


class TOMLFormat(EditFormat):
    """TOML file format"""

    def __init__(self):
        super().__init__("toml")

    def format_entry_for_edit(self, entry: TimeEntry) -> str:
        """Format a time entry as TOML for editing"""
        entry_dict = entry.model_dump(mode="json")
        # NOTE: including none_value="null" so we can show all fields in the editor
        # even if they are None, it's a bit sketchy but it works for now as long as
        # we don't have any fields that are actually the string "null". It's really
        # a workaround to the fact that the TOML spec doesn't support null values.
        return toml.dumps(entry_dict, none_value="null")

    def parse_edited_entry(self, content: str) -> TimeEntry:
        """Parse edited TOML content back into a TimeEntry"""
        try:
            # NOTE: see comment in format_entry_for_edit where we serialize None values
            # as "null". TOML spec doesn't support null values so we have to work around
            data = toml.loads(content, none_value="null")
            return TimeEntry(**data)
        except Exception as e:
            raise EditorError(f"Invalid entry format: {str(e)}") from e


class ShellEditor(Editor):
    """Editor for shelling to editor to update time entries"""

    def __init__(self, editor_command: str, format: EditFormat | str | None = None):
        super().__init__(editor_command)
        if isinstance(format, str) and format.lower() == "toml":
            self.format = TOMLFormat()
        elif isinstance(format, EditFormat):
            self.format = format
        else:
            self.format = YAMLFormat()

    def edit_entry(self, entry: TimeEntry) -> TimeEntry:
        """Edit a time entry in the user's preferred editor"""
        # Create temp file with entry content
        with tempfile.NamedTemporaryFile(suffix=self.format.suffix, mode="w+", delete=False) as tmp:
            tmp.write(self.format.format_entry_for_edit(entry))
            tmp.flush()
            tmp_path = tmp.name

        try:
            # Open editor
            subprocess.run([self.editor_command, tmp_path], check=True)

            # Read and parse edited content
            with open(tmp_path) as f:
                edited_content = f.read()
                updated_entry = self.format.parse_edited_entry(edited_content)

            return updated_entry

        finally:
            # Clean up temp file
            os.unlink(tmp_path)



from ...models import TimeEntry
from .. import CsvOutput


def test_csv_output_single_entry():
    entry = TimeEntry(
        id="1234567890",
        start_time="2021-01-01T00:00:00",
        end_time="2021-01-01T01:00:00",
        project="test",
        comment="test",
        tags=["test"],
    )
    output = CsvOutput()
    output.single_entry_output(entry)


def test_csv_output_multiple_entries():
    entries = [
        TimeEntry(
            id="1234567890",
            start_time="2021-01-01T00:00:00",
            end_time="2021-01-01T01:00:00",
            project="test",
            comment="test",
            tags=["test"],
        ),
        TimeEntry(
            id="1234567890",
            start_time="2021-01-01T00:00:00",
            end_time="2021-01-01T01:00:00",
            project="test",
            comment="test",
            tags=["test"],
        ),
    ]
    output = CsvOutput()
    output.multiple_entries_output(entries)


def test_csv_output_export():
    output = CsvOutput()
    output.export_output(2, "test.csv")


def test_csv_output_no_entry():
    output = CsvOutput()
    output.single_entry_output(None)



import csv
import sys
from typing import Any

from ..models import TimeEntry
from .output import OutputFormatter


class CsvOutput(OutputFormatter):
    def _get_header(self):
        return [
            "id",
            "start_time",
            "end_time",
            "project",
            "comment",
            "tags",
        ]

    def _get_row(self, entry: TimeEntry):
        return [
            entry.id,
            entry.naive_start_time.strftime("%Y-%m-%d %H:%M:%S"),
            entry.naive_end_time.strftime("%Y-%m-%d %H:%M:%S") if entry.naive_end_time else "",
            entry.project,
            entry.comment,
            "|".join(entry.tags),
        ]

    def _send_output(self, output: list[list[Any]]) -> None:
        writer = csv.writer(sys.stdout)
        writer.writerows(output)

    def single_entry_output(self, entry: TimeEntry | None) -> None:
        if entry is None:
            return
        self._send_output([self._get_header(), self._get_row(entry)])

    def multiple_entries_output(self, entries: list[TimeEntry]) -> None:
        self._send_output([self._get_header()] + [self._get_row(entry) for entry in entries])

    def export_output(self, count: int, filename: str) -> None:
        self._send_output([["exported", "filename"], [count, filename]])



from datetime import timedelta
from pathlib import Path

from jinja2 import Template
from pydantic import BaseModel

from ..models import TimeEntry
from ..utils.datetime_utils import format_delta
from .output import OutputFormatter

SINGLE_ENTRY = Template(
    """
# Time Entry

| field | value |
|-------|-------|
| ID | {{ entry.id }} |
| Start Time | {{ entry.naive_start_time.strftime("%Y-%m-%d %H:%M:%S") }} |
| Stop Time | {{ entry.naive_end_time.strftime("%Y-%m-%d %H:%M:%S") }} |
| Delta | {{ entry.humanized_duration }} |
| Project | {{ entry.project }} |
| Comment | {{ entry.comment }} |
| Tags | {{ entry.tags }} |
"""
)


MULTIPLE_ENTRIES = Template(
    """
# Time Entries

| ID | Start Time| Stop Time| Delta |Project |Comment |Tags |
|----|----------:|---------:|------:|--------|--------|-----|
{% for entry in entries %}{% if 'total_duration' in entry.model_fields_set %}| | {{ entry.total_description }}| | {{ entry.total_duration }} | | | |{% else %}| {{ entry.id[0:4] }} | {{ entry.naive_start_time.strftime("%Y-%m-%d %H:%M:%S") }}| {{ entry.naive_end_time.strftime("%H:%M:%S") }} | {{ entry.humanized_duration }}| {{ entry.project }} | {{ entry.comment }} | {{ ", ".join(entry.tags) }} |{% endif %}
{% endfor %}
"""  # noqa: E501
)


EXPORT_OUTPUT = Template(
    """
Exported {{ count }} entries to {{ filename }}
"""
)


class OutputModel(BaseModel):
    total_description: str = ""
    total_duration: str = ""


class MarkdownOutput(OutputFormatter):
    def single_entry_output(self, entry: TimeEntry | None) -> None:
        if entry is None:
            print("No active time record.")
        else:
            print(SINGLE_ENTRY.render(entry=entry))

    def multiple_entries_output(self, entries: list[TimeEntry]) -> None:
        output_entries = []
        if len(entries):
            current_date = entries[0].naive_start_time.date()
            subtotal_delta = timedelta(0)
            total_delta = timedelta(0)
            for entry in entries:
                total_delta += (entry.naive_end_time - entry.naive_start_time) if entry.naive_end_time else timedelta(0)
                if entry.naive_start_time.date() != current_date:
                    output_entries.append(
                        OutputModel(total_description="*subtotal*", total_duration=format_delta(subtotal_delta))
                    )
                    current_date = entry.naive_start_time.date()
                    subtotal_delta = timedelta(0)
                subtotal_delta += (
                    (entry.naive_end_time - entry.naive_start_time) if entry.naive_end_time else timedelta(0)
                )
                output_entries.append(entry)
            output_entries.append(
                OutputModel(total_description="*subtotal*", total_duration=format_delta(subtotal_delta))
            )
            output_entries.append(OutputModel(total_description="**total**", total_duration=format_delta(total_delta)))
        print(MULTIPLE_ENTRIES.render(entries=output_entries))

    def export_output(self, count: int, filename: Path | str) -> None:
        print(EXPORT_OUTPUT.render(count=count, filename=filename))



import io
import json

import pytest
import rtoml as toml
import ryaml

from ...models import TimeEntry
from ..time_entry_repo_file import FormatFactory, JSONFormat, TimeEntryRepositoryFile, TOMLFormat, YAMLFormat


def test_file_format_factory():
    factory = FormatFactory()
    assert isinstance(factory.get_format("yaml"), YAMLFormat)
    assert isinstance(factory.get_format("toml"), TOMLFormat)
    assert isinstance(factory.get_format("json"), JSONFormat)
    assert isinstance(factory.get_format("yml"), YAMLFormat)
    with pytest.raises(ValueError):
        factory.get_format("unknown")


def test_yaml_format():
    fmt = YAMLFormat()
    data = {"key": "value"}
    assert fmt.load_data(io.StringIO(ryaml.dumps(data))) == data
    with io.StringIO() as f:
        fmt.save_data(data, f)
        f.seek(0)
        assert ryaml.dumps(data) == f.read()


def test_toml_format():
    fmt = TOMLFormat()
    data = {"key": "value"}
    assert fmt.load_data(io.StringIO(toml.dumps(data))) == data
    with io.StringIO() as f:
        fmt.save_data(data, f)
        f.seek(0)
        assert toml.dumps(data) == f.read()


def test_json_format():
    fmt = JSONFormat()
    data = {"key": "value"}
    assert fmt.load_data(io.StringIO(json.dumps(data))) == data
    with io.StringIO() as f:
        fmt.save_data(data, f)
        f.seek(0)
        assert json.dumps(data) == f.read()


def test_supported_formats():
    assert FormatFactory.get_supported_formats() == ["yaml", "yml", "toml", "json"]


@pytest.mark.parametrize("fmt", FormatFactory.get_supported_formats())
def test_time_entry_repo_file(tmp_path, fmt):
    filename = tmp_path / f"test.{fmt}"
    repo = TimeEntryRepositoryFile(filename)

    assert repo.get_all() == []
    with pytest.raises(KeyError):
        repo.get_entry_by_id("1")

    entry = TimeEntry(
        project="test", start_time="2021-01-01T00:00:00", end_time="2021-01-01T01:00:00", comment="test comment"
    )
    repo.save(entry)

    assert repo.get_entry_by_id(entry.id) == entry

    entry2 = TimeEntry(project="test2", start_time="2021-01-02T00:00:00", comment="test comment 2")

    repo.save(entry2)
    assert repo.get_entry_by_id(entry2.id) == entry2

    repo.delete_entry(entry2.id)
    with pytest.raises(KeyError):
        repo.get_entry_by_id(entry2.id)

    with pytest.raises(KeyError):
        repo.delete_entry(entry2.id)

    assert repo.filter(filter=None) == [entry]

    assert repo.get_by_project("test") == [entry]
